/*
SEARCH32 -- Determine how fast data can be transferred from a hard drive or network.
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with Visual Studio 6 or later.  Requires comdlg32.lib odbc32.lib odbccp32.lib winmm.lib
	cl search32.c /link netapi32.lib odbc32.lib odbccp32.lib mpr.lib version.lib

*/


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <math.h>
#ifdef WIN32
#include <ctype.h>
#include <process.h>
#include <io.h>
#include <fcntl.h>
#include <windows.h>
#include <winver.h>
#include <aclapi.h>
#define achName cAlternateFileName
#include <direct.h>
#define DIRECTORY 0x10
#define ARCHIVED 0x20
#define HIDDEN   0x02
#define SYSTEM   0x04
#else
// For Linux
#define HANDLE long
#define _MAX_PATH PATHMAX
#endif
char ascdate[9];
unsigned filedate;
char fileform,hidden_system,archived,calculate_crc=0;
#define BUFSI 60000
#define MAX_EXCLUDES 255

char notempexcludes[]="\\TEMP\\\nCOOKIES\n.LNK\n.TMP\nTEMPORARY\n$NtUninstal";
int recursion = 1;
unsigned blocksize = 0;
unsigned block_size[]={
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
FILE *otfi;
time_t cfidate, cfidate_hi, afidate, afidate_hi; //write dates and access dates to check
char tflag=0,totalonly = 0,display_version_info = 1; 
char earliest_date = 0, latest_date = 0, largest_file = 0, smallest_file = 0;
char earliest_activity = 0, latest_activity = 0, first_file = 0;
_int64 total_space = 0, total_used = 0;
long total_files = 0;
_int64 sizelow,sizehigh;
_int64 smallest_size = 9999999999, largest_size = 0;
time_t early_date = 0x7fffffff, late_date = 0;
struct stat filestat;
char early_file[_MAX_PATH], late_file[_MAX_PATH], small_file[_MAX_PATH], large_file[_MAX_PATH];
int return_code = 1, search_string_count = 0, exclude_string_count = 0;
static char driveletter[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
char net_connection[_MAX_PATH][26]; // Save the network connection for each network drive
char logfile=0, vflag=1, quiet=0, *exclude_path[MAX_EXCLUDES], vplus = 0;
char search_string[16][256], pathflag=0, tabs = 0,space_or_tab[] = {' ',0};
char checkStreams = 0;

char short_name = 0;
FILE *logfi;
struct stat fistat;
char use_commas = 0;

#define LMAX	512
#define PMAX	512
#define ANY	    '?'
#define CLASS	'['
#define ENDCLASS ']'
#define ESCAPE  '\\'
#define STAR	'*'
#define RANGE '-'
#define ENDPAT  0
char	lbuf[LMAX];
char	pbuf[PMAX];
char	*pp;
int		file_version[4];


#define CRC32_POLY 0x04c11db7     /* AUTODIN II, Ethernet, & FDDI */
unsigned long crc32_table[256];

VOID *context;
HANDLE fileHandle;

WIN32_FIND_DATA dta;
WORD fdateLastAccess, fdateLastWrite, ftimeLastWrite, ftimeLastAccess, fdateCreate, ftimeCreate;

#ifdef WIN32
WIN32_STREAM_ID stream;

int isNTFS(char *drive) {
	// Return 1 if the file system on device i is NTFS, else 0
	char fileSystemName[64], VolumeNameBuffer[64];
	DWORD nFileSystemNameSize = 64;
	DWORD MaximumComponentLength;
	DWORD FileSystemFlags;

	
	GetVolumeInformation(drive, VolumeNameBuffer, 64, NULL, &MaximumComponentLength, &FileSystemFlags,
		fileSystemName, nFileSystemNameSize);
	return (strcmp(fileSystemName,"NTFS") ? 0 : 1);
}



// Close the File Handle and shut down Backup Read
int CloseBackupReadHandle(WCHAR *streamname) {
	DWORD BytesRead;
	stream.dwStreamAttributes=0;
	stream.dwStreamId=0;
	stream.dwStreamNameSize=0;
	stream.Size.QuadPart = 0;
	streamname[0] = 0;
	BackupRead(fileHandle,(unsigned char *)&stream,0,&BytesRead,TRUE,FALSE,&context);
	CloseHandle(fileHandle);
	return 0;
}


int FirstStream(char *filename,char *StreamNameAscii,_int64 *filesize,int *streamId) {
	WCHAR streamname[_MAX_PATH];
	DWORD BytesRead,actualLow,actualHigh,namesize;
	DWORD rc;
	int StreamHeaderSize = ((LPBYTE)stream.cStreamName - (LPBYTE)&stream);

	context=NULL;
	fileHandle = CreateFile(filename,GENERIC_READ,FILE_SHARE_READ,
			NULL,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,NULL);

	if (fileHandle == INVALID_HANDLE_VALUE) {
		rc = GetLastError();
		return 0;
	}

	rc = 1;
	BytesRead=1;
	stream.dwStreamAttributes=0;
	stream.dwStreamId=0;
	stream.dwStreamNameSize=0;
	stream.Size.QuadPart = 0;
	streamname[0] = 0;
	// Read the first 20 bytes of the file into stream
	// This gives the stream id, stream attributes,
	// 
	rc=BackupRead(fileHandle,(unsigned char *)&stream,StreamHeaderSize,&BytesRead,FALSE,FALSE,&context);
	if (rc && BytesRead) {
		namesize = stream.dwStreamNameSize;
		*filesize = stream.Size.QuadPart;
		*streamId = stream.dwStreamId;
		if (namesize) {
			// read the file name
			rc = BackupRead(fileHandle,(unsigned char *)streamname,namesize,
			&BytesRead,FALSE,FALSE,&context);
			namesize /= sizeof(WCHAR);
			streamname[namesize]=0;
		}
		
		sprintf(StreamNameAscii,"%ls",streamname);
		rc = BackupSeek(fileHandle,stream.Size.LowPart,stream.Size.HighPart,&actualLow,&actualHigh,&context);
		return 1;
	}
	
	CloseBackupReadHandle(streamname);
	
	return 0;
}

int NextStream(char *StreamNameAscii,_int64 *filesize,int *streamId) {

	WCHAR streamname[_MAX_PATH];
	DWORD BytesRead,actualLow,actualHigh,namesize;
	DWORD rc;
	int StreamHeaderSize = ((LPBYTE)stream.cStreamName - (LPBYTE)&stream);

	rc = 1;
	BytesRead=1;
	stream.dwStreamAttributes=0;
	stream.dwStreamId=0;
	stream.dwStreamNameSize=0;
	stream.Size.QuadPart = 0;
	streamname[0] = 0;
	// Read the first 20 bytes of the file into stream
	// This gives the stream id, stream attributes,
	// 
	rc=BackupRead(fileHandle,(unsigned char *)&stream,StreamHeaderSize,&BytesRead,FALSE,FALSE,&context);
	if (rc && BytesRead) {
		namesize = stream.dwStreamNameSize;
		*filesize = stream.Size.QuadPart;
		*streamId = stream.dwStreamId;
		if (namesize) {
			// read the file name
			rc = BackupRead(fileHandle,(unsigned char *)streamname,namesize,
			&BytesRead,FALSE,FALSE,&context);
			namesize /= sizeof(WCHAR);
			streamname[namesize]=0;
		}
		
		sprintf(StreamNameAscii,"%ls",streamname);
		rc = BackupSeek(fileHandle,stream.Size.LowPart,stream.Size.HighPart,&actualLow,&actualHigh,&context);
		return 1;
	}
	CloseBackupReadHandle(streamname);
	return 0;
}

int winFileVersion(char *path,unsigned file_version[4]) {
	BYTE   *abData;
	DWORD  handle;
	DWORD  dwSize;
	//	LPBYTE lpBuffer;
	VS_FIXEDFILEINFO *lpBuffer;
	//	Get File Version Info, if present 

	file_version[0]=0;
	file_version[1]=0;
	file_version[2]=0;
	file_version[3]=0;

	__try {
		dwSize = GetFileVersionInfoSize(path, &handle);
	}
	__except (1) {
		file_version[0] = -1;
		file_version[1] = -1;
		file_version[2] = -1;
		file_version[3] = -1;
		return 1;
	}
	if (dwSize > 0) {
		abData = malloc(dwSize);
		if (abData != NULL) {
			GetFileVersionInfo(path, handle, dwSize, abData);
			VerQueryValue(abData, "\\", &lpBuffer, &dwSize);
			if (dwSize!=0) {    
				//memcpy(&fileinfo,lpBuffer,sizeof(fileinfo));
				file_version[0] = HIWORD(lpBuffer->dwFileVersionMS);
				file_version[1] = LOWORD(lpBuffer->dwFileVersionMS);
				file_version[2] = HIWORD(lpBuffer->dwFileVersionLS);
				file_version[3] = LOWORD(lpBuffer->dwFileVersionLS);
				free (abData);
				return 1;
			}
			free (abData);
			return 0;
		}
		return 0;
	}
	return 0;
}

void fidtetoa(ascdate,filedate)
char *ascdate;
unsigned filedate;
{
	unsigned mm,dd,yy;
	char temp[9];
	yy = (filedate>>9) + 1980;
	yy %= 100;
	dd = filedate&0x1f;
	mm = (filedate & 0x01e0) >> 5;
	/* convert to ascii and store as mm/dd/yy */
	sprintf(temp,"%2d/%2d/%2d",mm,dd,yy);
	strncpy(ascdate,temp,8);
	if (ascdate[0]==' ') ascdate[0] = '0';
	if (ascdate[3]==' ') ascdate[3] = '0';
	if (ascdate[6]==' ') ascdate[6] = '0';
}

time_t dtetim_t(filedate,filetime)
unsigned filedate,filetime;
{	
	struct tm fitm;

	fitm.tm_year = (filedate>>9) + 80; /* Years since 1900 */
	fitm.tm_mday = filedate&0x1f;
	fitm.tm_mon = ((filedate & 0x01e0) >> 5) - 1;
	fitm.tm_hour = filetime>>11;
	fitm.tm_isdst = -1;
	if (fitm.tm_hour > 23)
		fitm.tm_hour = 23;
	fitm.tm_sec =  filetime&0x1f; /* seconds in 2 second intervals */
	fitm.tm_sec += fitm.tm_sec;	   /* seconds in 1 second intervals */
	if (fitm.tm_sec > 59)
		fitm.tm_sec = 59;
	fitm.tm_min = (filetime & 0x07e0) >> 5;
	if (fitm.tm_min > 59)
		fitm.tm_min = 59;
   	/* convert to seconds since 1970 */
	return mktime(&fitm);
}

typedef struct streaminfo {
		_int64 streamSize;
		struct streaminfo *streamNext;
		int streamId;
		char streamName[_MAX_PATH];
} STREAMLIST;

STREAMLIST *buildStreamList(char *filename, _int64 *total_space) {
// Build a linked list of streams, return top of list on exit
	int streamId, rc, mallocsize = sizeof(STREAMLIST);
	STREAMLIST *firststream=NULL, *nextstream, *current;
	char streamname[_MAX_PATH], *cdata;
	_int64 filesize;
	*total_space = 0;
	rc = FirstStream(filename,streamname,&filesize,&streamId);
	while (rc) {
		/* found the first stream in the file */
		/* Allocate space for a new stream entry */
		nextstream = (STREAMLIST *) malloc(mallocsize);
		if (nextstream == NULL) {
			fprintf(stderr,"Allocation Error.  Unable to continue.\n");
			exit(8);
		}
		/* Fill in the data returned from the stream function */
		nextstream->streamSize = filesize;
		nextstream->streamId = streamId;
		*total_space += filesize;
		nextstream->streamNext = NULL;
//		strcpy(nextstream->streamName, filename);
		// Remove :$DATA from the stream name
		cdata = strstr(streamname, ":$DATA");
		if (cdata)
			*cdata = 0;
		strcpy(nextstream->streamName, streamname);
		if (firststream) 
			current->streamNext = nextstream;
		else
			firststream = nextstream;
		current=nextstream;
		rc = NextStream(streamname,&filesize,&streamId);
	}
	return firststream;
}

// capture file attributes
char *GetFileAttrib(char *buf) {
	*buf = 0;
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE ? "A" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_COMPRESSED ? "C" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ? "D" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_ENCRYPTED ? "E" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_HIDDEN ? "H" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_OFFLINE ? "O" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_READONLY ? "R" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT ? "P" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_SPARSE_FILE ? "W" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_SYSTEM ? "S" : "-");
	strcat(buf,dta.dwFileAttributes & FILE_ATTRIBUTE_TEMPORARY ? "T" : "-");
	return buf;
}


#else
// Used under linux.  Dummy out the above routines
int isNTFS(char *drive) {
	return 0;
}

// Close the File Handle and shut down Backup Read
int CloseBackupReadHandle(WCHAR *streamname) {
	return 0;
}

int FirstStream(char *filename,char *StreamNameAscii,_int64 *filesize,int *streamId) {
	return 0;
}

int NextStream(char *StreamNameAscii,_int64 *filesize,int *streamId) {
	return 0;
}

int winFileVersion(char *path,unsigned file_version[4]) {
	return 0;
}

// rewrite these routines for Unix file times
void fidtetoa(ascdate,filedate)
char *ascdate;
unsigned filedate;
{
	unsigned mm,dd,yy;
	char temp[9];
	yy = (filedate>>9) + 1980;
	yy %= 100;
	dd = filedate&0x1f;
	mm = (filedate & 0x01e0) >> 5;
	/* convert to ascii and store as mm/dd/yy */
	sprintf(temp,"%2d/%2d/%2d",mm,dd,yy);
	strncpy(ascdate,temp,8);
	if (ascdate[0]==' ') ascdate[0] = '0';
	if (ascdate[3]==' ') ascdate[3] = '0';
	if (ascdate[6]==' ') ascdate[6] = '0';
}

time_t dtetim_t(filedate,filetime)
unsigned filedate,filetime;
{	
	struct tm fitm;

	fitm.tm_year = (filedate>>9) + 80; /* Years since 1900 */
	fitm.tm_mday = filedate&0x1f;
	fitm.tm_mon = ((filedate & 0x01e0) >> 5) - 1;
	fitm.tm_hour = filetime>>11;
	fitm.tm_isdst = -1;
	if (fitm.tm_hour > 23)
		fitm.tm_hour = 23;
	fitm.tm_sec =  filetime&0x1f; /* seconds in 2 second intervals */
	fitm.tm_sec += fitm.tm_sec;	   /* seconds in 1 second intervals */
	if (fitm.tm_sec > 59)
		fitm.tm_sec = 59;
	fitm.tm_min = (filetime & 0x07e0) >> 5;
	if (fitm.tm_min > 59)
		fitm.tm_min = 59;
   	/* convert to seconds since 1970 */
	return mktime(&fitm);
}

// capture file attributes Unix
char *GetFileAttrib(char *buf) {
	*buf = 0;
	// rewrite for Unix using definitions for stat.st_mode
	return buf;
}

#endif

unsigned long
updcrc32(unsigned long icrc, unsigned char *icp, unsigned int icnt) {
	register unsigned long crc = icrc;
	register unsigned char *cp = icp;
	register unsigned int cnt = icnt;

	while ( cnt-- ) {
		crc = (crc<<8) ^ crc32_table[(crc>>24) ^ *cp++];
	}

	return( crc );
}

void init_crc32()
{
        int i, j;
        unsigned long c;

        for (i = 0; i < 256; ++i) {
                for (c = i << 24, j = 8; j > 0; --j)
                        c = c & 0x80000000 ? (c << 1) ^ CRC32_POLY : (c << 1);
                crc32_table[i] = c;
        }
}

unsigned long calculate_crc32(char *filenm,_int64 filesize) {
	unsigned long crc32;
	int l;
	int infi;
	char buffer[BUFSI];

	crc32 = (unsigned) (0xFFFFFFFF);
	if (filesize == 0)
		return crc32;

	if ((infi = open(filenm,O_RDONLY+O_BINARY)) < 1) {
		fprintf(stderr,"Unable to open ");
		perror(filenm);
		return crc32;
	}

	do {
		l = read(infi,buffer,(unsigned)(BUFSI));
		crc32 = updcrc32(crc32,buffer,l);
	}
	while (l == BUFSI);
	
	close(infi);
	return crc32;
}



uxmatch(char *line, char *pattern)

/* (partial) line to match	   */
/* (partial) pattern to match	 */
{
	register char   *l;	    /* Current line pointer	      */
	register char   *p;	    /* Current pattern pointer      */
	int		   op;	      	/* Pattern operation	      */
	char inclass = 0, lowvalue = 0, foundmatch = 0, escape = 0;
	l = line;
	p = pattern;
	if (*p == ENDPAT)
		return 1;

	while ((op = *p++) != ENDPAT) {
		escape = 0;
		if (op == '\\') {
			op = *p++;
			if (op)		
				escape = 1;
			else {	/* pattern ends with \. look for \ */
				--p;
				op = '\\';
			}
		}

		if (inclass) {
			/* currently scanning a class spec [abc-de] */
			if (!foundmatch) {
				/* and we haven't found a match yet */
				if (lowvalue) {
				/* but we are in a range c-e */
				/* op is high value of range */
					if (tolower(*l) >= lowvalue && tolower(*l) <= tolower(op)) {
						/* we found a match */
						foundmatch = 1;
						lowvalue = 0;
						continue;
					}
				/* no match found, return failure */
					lowvalue = 0;
				}
				else {
					/* not in a range */
					if (*p == RANGE)  {
						/* in a range, op is low value */
						lowvalue = (char) tolower(op);
						++p;
						continue;
					}
					else
						lowvalue = 0;

					if (tolower(*l) == tolower(op)) {
						/* found match, whoopie */
						foundmatch = 1;
						continue;
					}
				}
			}
			if (op == ENDCLASS && !escape) {
				if (!foundmatch)					
					return 0;		/* not found, sorry charlie */
				inclass = 0;
				++l;
			}
		}
		else {
			switch (op) {
			case ANY:
				if (*l++ == '\0')
					return(0);
				break;
			case CLASS:
				inclass = 1;
				lowvalue = 0;
				foundmatch = 0;
				break;				
			case STAR:		 /* Zero or more characters	*/
																																										/* match line against rest of pattern */
				if (*l == 0 && *p ==0)
					// Currently looking for *, 0 or more characters
					// The strings is empty
					// If the pattern ends with * then success; else failure
					return 1;
				while (*l)
					if (uxmatch(l++,p))
						return 1;
				return 0;		 /* Nothing else worked */
			default:
				if (tolower(*l) != tolower(op))
					return(0);
				l++;
				break;
			}
		}
	}
	return *l == 0 ? 1 : 0;   /* if everything matches and at end of string, ok */
}

char *comma_delim(_int64 lno,char *buffer,char mb,char kb,char use_commas) {
/* returns the long integer in comma delimited format */
/* Assume buffer is long enough to hold the largest value */
	char b1[36],*b; /* should be long enough to hold largest ltoa value */
	int i,n,l,k,c;
	_int64 no;
	no = lno;
	if (mb) {
		no = (lno / (1024*1024));
		if (no > -1.0 && no < 1.0) {
			kb = 1;
			mb = 0;
		}
	}

	if (kb) {
		no = (lno / 1024);
		if (no > -1.0 && no < 1.0) {
			no = lno;
			kb = 0;
		}
	}

	sprintf(b1,"%16I64d",no);
	for (b=b1;*b==' ';++b) ;  // Remove leading blanks
	if (use_commas == 0) {
		strcpy(buffer,b);
		return buffer;
	}

	l = strlen(b); /* l is length of string and index to 0 at end of string */
	n = l/3;	   /* n is number of commas to be inserted */
	if ((l % 3) == 0)
		--n;
	k = l + n;	   /* k is length of new string */
	buffer[k--] = b[l--];
	for (c=0,i=l; i >= 0; --i,--k) {
		buffer[k] = b[i];
		if (++c == 3 && i > 0) {
			buffer[--k] = ',';
			c = 0;
		}
	}

	if (mb)
		strcat(buffer,"MB");
	else if (kb)
		strcat(buffer,"KB");

	return buffer;
}

void fidatetime(char *asctime, unsigned filedate, unsigned filetime) {
	char temp[9];
	unsigned mm,dd,yy,hh,ss;
//  format date as yyyymmdd
	yy = (filedate>>9) + 1980;
	dd = filedate&0x1f;
	mm = (filedate & 0x01e0) >> 5;
	sprintf(temp,"%4d%2d%2d",yy,mm,dd);
	if (temp[4]==' ') temp[4]='0';
	if (temp[6]==' ') temp[6]='0';
	strcpy(asctime,temp);
//  format time as hhmmss
	hh = filetime>>11;
	ss = filetime&0x1f;
	ss = ss + ss;
	mm = (filetime & 0x07e0) >> 5;
	sprintf(temp,"%2d%2d%2d",hh,mm,ss);
	if (temp[0] == ' ') temp[0] = '0';
	if (temp[2] == ' ') temp[2] = '0';
	if (temp[4] == ' ') temp[4] = '0';
	strcat(asctime, temp);
}


void fitimtoa(asctime,filetime)
char *asctime;
unsigned filetime;
{
	int hh,mm,ss;
	char temp[9];
	hh = filetime>>11;
	ss = filetime&0x1f;
	ss = ss + ss;
	mm = (filetime & 0x07e0) >> 5;
/* convert to ascii and store as hh:mm:ss */
	sprintf(temp,"%2d:%2d:%2d",hh,mm,ss);
	strncpy(asctime,temp,8);
	if (asctime[0] == ' ') asctime[0] = '0';
	if (asctime[3] == ' ') asctime[3] = '0';
	if (asctime[6] == ' ') asctime[6] = '0';

}


int file_contains(
	char *filename,
	int searchkt,
	char searchstring[16][256],
	_int64 filesize,
	unsigned long *crc32) {
	// open filename, look for searchstring, return 1 if found; else 0
	int f;
	char buf[65535], temp[_MAX_PATH];
	int len,slen[16],i,ln, offset, k;

	*crc32 = (unsigned) (0xFFFFFFFF);

	for (k=0;k<searchkt;++k)
		slen[k] = strlen(searchstring[k]);

	f = open(filename,_O_BINARY | _O_RDONLY);
	if (f == -1) {
		sprintf(temp,"Unable to open file: %s.  ",filename);
		perror(temp);
		return 0;
	}	
	
	offset = 0; // read in first buffer at start of buf
	
	while (len = read(f,buf+offset,sizeof(buf)-offset)) {
		if (len < 0 || len > sizeof(buf)) {
			perror("Error reading file.");
			close(f);
			return 0;
		}
		ln = len;
		if (calculate_crc)
			*crc32 = updcrc32(*crc32,buf+offset,len);

		for (k=0; k< searchkt; ++k) {
			for (i=0;i < (len - slen[k]); ++i) {
				if (memicmp(buf+i,searchstring[k],slen[k]) == 0) {
					close(f);
					return 1;
				}
			}
		}
		memcpy(buf, buf+len-256, 256);
		offset = 256;
	}
	close(f);
	return 0;
}

// create full name with path info from a name and path
char fullnamepath[_MAX_PATH];
char *fullname(char *path,char *name) {
	char temp[_MAX_PATH];
	strcpy(temp,path);
	if (temp[3]==0) temp[2] = 0;
	strcat(temp,"\\");
	if (tabs)
		strcat(temp,"\t"); 
	// If tabs set, then put a tab between filepath and
	// file name for easier parsing.
	strcat(temp,name);
	strcpy(fullnamepath,temp);
	return fullnamepath;

}

// check to see if the file should not be searche
int file_is_excluded(int exclude_string_count,char *path) {
	int l,excluded;

	strupr(path);
#ifdef UNIX_EXCLUDES
	// This code implements exclusion using Unix style wild cards
	// The logic is simple, but the specification is difficult
	// This give a more fine grain control, but you have to be very
	// careful in specifying the exclude path
	// The exclude path can contain Unix style wild card characters.
	if (exclude_string_count) {
		for (excluded=0,l = 0; l < exclude_string_count && !excluded; ++l) {
			if (uxmatch(path,exclude_path[l]))
				excluded = 1;	// skip this file
		}
		if (excluded)
			return 1;
	}
#else
	// This code implements exclusion using simple string inclusion
	// If the selected file contains any one of the exclude strings
	// the file is excluded.  Under this option, you may not specify
	// Wildcard characters and the \ is not a special character
	if (exclude_string_count) {
		for (excluded=0,l = 0; l < exclude_string_count && !excluded; ++l) {
			if (strstr(path,exclude_path[l]))
				excluded = 1;	// skip this file
		}
		if (excluded)
			return 1;
	}
#endif			
	return 0;
}
			
int EarlyLateBigSmall(char *path, time_t wr_dt, _int64 total_filesize) {
	// Depending on flags, set earliest, latest dates and names, largest, smallest dates and names
	int chg = 0;

	if (earliest_date) {
		if (early_date > wr_dt) {
			early_date = wr_dt;
			strcpy(early_file,path);
		}
		chg = 1;
	}

	if (latest_date) {
		if (late_date < wr_dt) {
			late_date = wr_dt;
			strcpy(late_file,path);
		}
		chg = 1;
	}

	if (smallest_file) {
		if (smallest_size > total_filesize) {
			smallest_size = total_filesize;
			strcpy(small_file,path);
		}
		chg = 1;
	}

	if (largest_file) {
		if (largest_size < total_filesize) {
			largest_size = total_filesize;
			strcpy(large_file,path);
		}
		chg = 1;
	}

	return chg;
}


char blanks[]="                                                                              ";

int blankpathdisplay(char *path_displayed) {
	/* If displaying each path as it is being searched, set display indicator */
	if (otfi == stdout && vflag && *path_displayed && isatty(_fileno( stdout ) )) {
		*path_displayed = 0;
		fprintf(stderr,"\r%s\r",blanks);
		return 1;
	}
	return 0;
}

/* print the data to otfi and logfi */
void plog(char *b, char *path_displayed) {
	/* if totalonly flag set, don't print anything; else print it now */
	/* If printing filenames only or this is the 2nd item on a row, print a linefeed */
	if (totalonly)
		return;

	/* If displaying each path as it is being searched, set display indicator */
	if (path_displayed)
		blankpathdisplay(path_displayed);

	if (b == NULL)
		return;

	fprintf(otfi,b);
	if (logfile)
		fprintf(logfi,b);
}




/* print the file information */
void printFileInfo(char *pth, char *current_filename, char *first_time_this_path, char *path_displayed, 
				   _int64 filesize, unsigned long crc32, char *streamname, int streamid) {
	
// Print the file information
// the first stream is filename, subsequent streams are filename:streamname

	char printline[_MAX_PATH], *logbuf, path[_MAX_PATH];
	static char oldpath[_MAX_PATH] = {0};
	static char StreamType[][10] = {"INVALID","DATA","EA_DATA","SECURITY","ALT_DATA","LINK","PROPERTY","OBJ_DATA","REPARSE","SPARSE"};
	char temp[64], file_attributes[12] = "";
	int i;
	
	SECURITY_INFORMATION SecurityInfo = OWNER_SECURITY_INFORMATION;
	PSID ppsidOwner, ppsidGroup;
	PACL ppDacl, ppSacl;
	PSECURITY_DESCRIPTOR ppSecurityDescriptor;
	DWORD rc=0, NameLen = 255, DomainNameLen = 255;
	char Name[255], DomainName[255];
	SID_NAME_USE peUse;
	
	logbuf = malloc(_MAX_PATH + 1024);

	strcpy(path,pth);
	strcpy(printline,current_filename); 

	if (display_version_info)
		winFileVersion(printline,file_version);

	/* if print filename only, then do it */
	if (fileform) {
		strcat(printline,streamname);
		strcat(printline,"\n");
		plog(printline, path_displayed);
		return;
	}

	/* format the line with the filename and needed data */

	/* If there is a new path, set up for display unless displaying the path as part of the name */
	if (strcmp(path,oldpath) && *first_time_this_path == 0 && !pathflag) {
		sprintf(logbuf,"%s\n",path);
		plog(logbuf,path_displayed);
		/* save the new path name */
		strcpy(oldpath,path);
		*first_time_this_path = 1; /* force directory display only once */
	}

	/* print a space or a nothing if the path name is part of the file name */
	plog(pathflag ? NULL : " ",NULL);

	if (totalonly)
		return;

	/* if printing crc, then don't print the short name */
	if (calculate_crc) 
		sprintf(printline,"%8x",crc32);
	else {
		/* no crc, print the short name or stream# */
		if (streamid !=1 )
			sprintf(printline,"  %9s",StreamType[streamid]);
		else {
			strcpy(printline,(dta.cAlternateFileName[0] ? 
				dta.cAlternateFileName : dta.cFileName));
			strupr(printline);
		}
	}
	
	/* print the filename or crc and filesize */
	if (tabs) 
//		sprintf(logbuf,"%s\t%.%I64d",printline,filesize);
		sprintf(logbuf,"%s\t%s",printline,comma_delim(filesize,temp,0,0,use_commas));
	else 
		sprintf(logbuf,"%-12s %16s",printline,comma_delim(filesize,temp,0,0,use_commas));
	strcat(logbuf,space_or_tab);

	printline[8] = 0;
	fidtetoa(printline,fdateLastWrite);			/* add the date of last write */
	strcat(logbuf,printline);
	strcat(logbuf," ");
	fitimtoa(printline,ftimeLastWrite);			/* add the file time */
	printline[8] = 0;
	strcat(logbuf,printline);

	if (vplus | tabs) {
		// Add date and time of last access
		strcat(logbuf,space_or_tab);
		printline[8] = 0;
		fidtetoa(printline,fdateLastAccess);			/* add the date of last ACCESS */
		strcat(logbuf,printline);
		strcat(logbuf," ");
		fitimtoa(printline,ftimeLastAccess);			/* add the file time  LAST ACCESS*/
		printline[8] = 0;
		strcat(logbuf,printline);
		// Add date and time of file create
		printline[8] = 0;
		fidtetoa(printline,fdateCreate);
		strcat(logbuf,space_or_tab);
		strcat(logbuf,printline);
		strcat(logbuf," ");
		fitimtoa(printline,ftimeCreate);			/* add the file time  LAST ACCESS*/
		printline[8] = 0;
		strcat(logbuf,printline);
	}
	
	
	/* add the version info if flag is set */			
	if (display_version_info) {
		/* stick in a delimiter either a tab or a label */
		if (tabs)
			strcat(logbuf,"\t");
		else
			strcat(logbuf," Ver: ");
		/* format the version info */
		if (file_version[0] == -1 && file_version[1] == -1 &&
			file_version[2] == -1 && file_version[3] == -1)
			strcpy(printline,"**INVALID**");
		else
			sprintf(printline,"%d.%d.%d.%d",
			file_version[0],file_version[1],
			file_version[2],file_version[3]);

		if (strlen(printline) < 11 && !tabs) {
			strcat(printline,"           ");
			printline[11] = 0;
		}
		strcat(logbuf,printline);
	}

	i = 0; //offset into pathname to display path information
	if (vplus | tabs) {
		// Print file Attributes
		strcat(logbuf, space_or_tab);
		strcat(logbuf, GetFileAttrib(file_attributes));
		
		// Print file Owner and file name on a new line
//		strcat(logbuf,"\n\t");
		strcat(logbuf, space_or_tab);
	
		rc=GetNamedSecurityInfo(current_filename, SE_FILE_OBJECT, SecurityInfo,
		&ppsidOwner, &ppsidGroup, &ppDacl, &ppSacl, &ppSecurityDescriptor);

		rc = LookupAccountSid(NULL,ppsidOwner,Name, &NameLen, DomainName, &DomainNameLen, &peUse);
		strcat(logbuf,DomainName);
		strcat(logbuf, "\\");
		strcat(logbuf, Name);
		strcat(logbuf, space_or_tab);
//		Print Machine Name
		if (path[0] == '\\' && path[1] == '\\') {
			// Have UNC name, extract servername.  Put machine name into temp
			for (i=2;path[i] && path[i] != '\\' && path[i] != '/';++i)
				temp[i-2] = path[i];
			temp[i-2] = 0;
			strcat(logbuf, temp);
			++i;	// skip over \ or /
			if (path[i+1] == '$')
				path[i+1] = ':';	// UNC Path points to a drive, display as a drive
			else 
				i = 0;				// UNC Path points to a share, display as UNC
		}
		else {
			GetEnvironmentVariable("COMPUTERNAME",temp, sizeof(temp));
			strcat(logbuf,temp );
		}
		strcat(logbuf, space_or_tab);
	}

	/* put in either the full pathname or the full file name */
	if (pathflag) {
		strcpy(temp, fullname(path+i,dta.cFileName));
		strcat(logbuf,temp);
	}
	else 
		strcat(logbuf,dta.cFileName);
	if (streamid)
		strcat(logbuf,streamname);
	strcat(logbuf,"\n");
	plog(logbuf,NULL);
	return;
}

char found_first_file = 0;
void look4(char *name[],int firstfile,int lastfile,char *path) {
	int k,rc,current_file,l;
	char pathdir[_MAX_PATH],temp[_MAX_PATH],oldpath[_MAX_PATH], volpath[6];
	char current_filename[_MAX_PATH];
	char searching_all = 0;
	HANDLE path_handle;
	unsigned short nummatch=1;
	unsigned long crc32;
	char driv[4], excluded = 0;
	time_t wr_dt, rd_dt;
	long longtemp = 0;
	_int64 filesize, filesize1, total_filesize;

	FILETIME localtime;
	int i, FileSystemNTFS;
	char chg, path_displayed = 0, first_time_this_path = 0;

	STREAMLIST *streams=NULL, myStream, *currentStream=NULL, *previousStream=NULL;

	unsigned long sectors_per_cluster,bytes_per_sector,total_clusters, free_clusters;

	if (first_file)
		if (found_first_file) // for this to be true, first_file must be true and a file found
			return;

	oldpath[0] = 0;
	longtemp = 0;
	if (vflag) {
		fprintf(stderr,"\r%s",blanks);
		if (strlen(path) > 64) {
			strncpy(temp,path,30);
			temp[30] = 0;
			strcat(temp,"...");
			strcat(temp,path + strlen(path) - 30);
		}
		else
			strcpy(temp,path);
		fprintf(stderr,"\rSearching: %s",temp);
		path_displayed = 1;
	}

	// If the currently selected directory path contains one of the strings in 
	// exclude_path don't report on this file.  Check before you look for files
	// If an exclude matches the file would be excluded anyway.  This will save
	// lots of time
	if (file_is_excluded(exclude_string_count,path))
		return;

	for (current_file=firstfile;current_file<=lastfile;++current_file) {
		strcpy(pathdir,path);
		if (path[3] == 0)
			pathdir[2] = 0;
		/* find current directory, drive */
		if (pathdir[strlen(pathdir)-1] != '\\')
			strcat(pathdir,"\\");
		if (strchr(name[current_file],'*') != NULL ||
		    strchr(name[current_file],'[') != NULL ||
		    name[current_file][0] == 0) {
			searching_all = 1;
			strcat(pathdir,"*.*");
		}
		else {
			strcat(pathdir,name[current_file]);
			searching_all = 0;
		}
		nummatch = 1;
		longtemp = 0;
		strncpy(volpath,pathdir,3);
		volpath[3] = 0;
		FileSystemNTFS = isNTFS(volpath);

	/* look for all files, including hidden and system */
		for (rc=TRUE,first_time_this_path=0,path_handle=FindFirstFile(pathdir,&dta);
         path_handle != INVALID_HANDLE_VALUE &&
         rc == TRUE;
         rc=FindNextFile(path_handle,&dta)) {

            if (first_file)
				if (found_first_file)
					return;

			if (dta.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
				continue; // ignore subdirectories, get all files first

			if (hidden_system && !(dta.dwFileAttributes & 
			    (FILE_ATTRIBUTE_HIDDEN+FILE_ATTRIBUTE_SYSTEM)))
				continue;

			if (archived && !(dta.dwFileAttributes & FILE_ATTRIBUTE_ARCHIVE))
				continue;

			FileTimeToLocalFileTime(&dta.ftLastWriteTime,&localtime);
			FileTimeToDosDateTime(&localtime,&fdateLastWrite,&ftimeLastWrite);
			wr_dt = dtetim_t(fdateLastWrite,ftimeLastWrite);


			FileTimeToLocalFileTime(&dta.ftLastAccessTime,&localtime);
			FileTimeToDosDateTime(&localtime,&fdateLastAccess,&ftimeLastAccess);
			rd_dt = dtetim_t(fdateLastAccess, ftimeLastAccess);


			FileTimeToLocalFileTime(&dta.ftCreationTime,&localtime);
			FileTimeToDosDateTime(&localtime,&fdateCreate,&ftimeCreate);
			// if (logfi) fprintf(logfi,"%s: %d %d %d\n",dta.cFileName,wr_dt,cfidate,cfidate_hi);

			if (wr_dt < cfidate || wr_dt > cfidate_hi)
				continue;
			if (rd_dt < afidate || rd_dt > afidate_hi)
				continue;

			filesize = ((_int64)(dta.nFileSizeHigh * 65536.) * (_int64)65536.) + (_int64)dta.nFileSizeLow;

			if (filesize < sizelow || filesize > sizehigh)
				continue;

			if (searching_all) {
				l = strlen(dta.cFileName);
				if (strchr(dta.cFileName,'.') == NULL) {
					dta.cFileName[l] = '.';
					dta.cFileName[l+1] = 0;
				}
				if (name[current_file][0] &&
				    ! uxmatch(dta.cFileName,name[current_file]))
					continue;
				dta.cFileName[l] = 0;
			}


			// place the filename into temp
			strcpy(temp,path);
			if (temp[3]==0) temp[2] = 0;
			strcat(temp,"\\");
			strcat(temp,((short_name && dta.cAlternateFileName[0]) ? dta.cAlternateFileName : dta.cFileName));				
			strcpy(current_filename, temp);

			// If the currently selected file name contains one of the strings in 
			// exclude_path don't report on this file.  
			if (file_is_excluded(exclude_string_count,current_filename))
				continue;

			// Gather non stream specific information, blocksize, free space, ...
			if (path[1] == ':') {
				i = path[0] -'@';
				if (block_size[i])
					blocksize = block_size[i];
				else {
					strncpy(driv,path,3);
					driv[3] = 0;
					GetDiskFreeSpace(
						driv,
						&sectors_per_cluster,
						&bytes_per_sector,
						&free_clusters,
						&total_clusters);

					blocksize = sectors_per_cluster * bytes_per_sector;
					block_size[i] = blocksize;
				}
			}

			// Build a linked list of stream names and filesizes in streams
			currentStream = NULL;
			previousStream= NULL;

			if (isNTFS && checkStreams) 
				// create a linked list of stream names and filesizes
				// The total filesize is the sum of the individual file sizes
				// The space used on disk is the total_filesize rounded to a
				// full block
				streams = buildStreamList(current_filename,&total_filesize);
			else {
				// Only one stream, create a list of a single entry
				streams = &myStream;
				*myStream.streamName = 0;
				myStream.streamSize = filesize;
				myStream.streamNext = NULL;
				myStream.streamId = 1;
				total_filesize = filesize;
			}

			// gather information on earliest date, latest date, smallest file, largest file
			chg = EarlyLateBigSmall(current_filename, wr_dt,total_filesize);

			/* if (chg)			Originally would stop, now let it continue
				continue;														*/
				

			total_files++;
			total_space += total_filesize;

			// If necessary, round up filesize to a complete block
			if (path[1] == ':') {
				_int64 blocks;
				blocks = total_filesize / blocksize;
				if (total_filesize % blocksize)
					++blocks;
				filesize1 = blocks * blocksize;
				total_used += filesize1;
			}
			else
				total_used += total_filesize;
						
			// for each stream in streams, gather info and print
			for (previousStream=NULL, currentStream = streams; currentStream; 
				previousStream=currentStream, currentStream=currentStream->streamNext) {

				if (previousStream)
					free(previousStream);

				strcpy(temp,current_filename);
				strcat(temp,currentStream->streamName);

				if (search_string_count) {
					// search file for string
					if (!file_contains(temp,search_string_count,search_string,filesize,&crc32))
						continue;
				}

				if (calculate_crc && (search_string_count == 0))
					crc32 = calculate_crc32(temp, filesize);

				printFileInfo(path, current_filename, &first_time_this_path, 
					&path_displayed, currentStream->streamSize, crc32, 
					currentStream->streamName, currentStream->streamId);

				if (first_file) {
					// done
					found_first_file = 1;
					break;
				}
			
			}
			
			nummatch = 1;
		}
		FindClose(path_handle);
		if (previousStream)
			if (previousStream != &myStream)
				free(previousStream);
		if (found_first_file) {
			return;
		}


	}
		

	longtemp = 0;
	strcpy(temp,path);
	if (temp[3] == 0) temp[2] = 0;
	strcat(temp,"\\*.*");
	for (k=0,rc = TRUE,path_handle=FindFirstFile(temp,&dta);
         path_handle != INVALID_HANDLE_VALUE &&
         rc == TRUE;
         rc=FindNextFile(path_handle,&dta)) {
		    if (!(dta.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
			continue;  // Directories Only
		    strcpy(pathdir,path);
		if (path[3] == 0)
			pathdir[2] = 0;
		if (pathdir[strlen(pathdir)-1] != '\\')
			strcat(pathdir,"\\");
		if (short_name)
			if (dta.cAlternateFileName[0])
				strcat(pathdir,dta.cAlternateFileName);
			else
				strcat(pathdir,dta.cFileName);
		else
			strcat(pathdir,dta.cFileName);
		if (dta.cFileName[0] != '.' && 
		    (dta.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) 
		    && recursion)
			look4(name,firstfile,lastfile,pathdir);
		nummatch = 1;
	}
	FindClose(path_handle);
	blankpathdisplay(&path_displayed);
}

/* extract a date from the parameter string */
time_t getadate(char *temp,char *actdate,char flag) {
/* date is of the form mmddyy[hh[mm[ss]]] or */
/*               (n),	meaning today + n days */
/*				 (nd),	meaning today + n years */
/*				 (nm),	meaning today + n months */
/*				 (nh), meaning now + n hours */

	int maxint=0x7fffffff;
	int maxsecs = 0x7fffffff/(24*3600);
	double fd_temp;
	char *paren,*c;
	char tmpp[65];
	long ndays,nmonths;
	char days[] = {0,31,29,31,30,31,30,31,31,30,31,30,31};
	unsigned cfidate;
	time_t today;
	double offset;
	struct tm *tm,fitm;
	int mm,dd,yy,k,hour,min,sec,start,gooddate,rc;
	cfidate = 0;
	hour=min=sec=0;
	if (flag != ' ') {
		hour = 23;
		min = 59;
		sec = 59;
	}
	if (temp[0] == '(') {
		// Time is Relative
		today = time(&today);
		paren = strchr(temp,')');
		ndays = atol(temp+1);
		tm = localtime(&today);
		for (c=temp+1;strchr("0123456789-+ ",*c);++c) ;
		switch(*c) {
		/* add number of days,months or years to todays date */
		/* if a negative number, it gives days before */
		/* today.  If Positive it gives days after today */
		case 'y':
		case 'Y':
			mm = tm->tm_mon + 1;
			dd = tm->tm_mday;
			yy = (int) tm->tm_year;
			yy = yy + ndays;	// Add in needed years
			if (yy > 121) {
				fprintf(stderr,"Invalid Date.  Date will be set to 1/1/2021\n");
				yy = 121;
				mm = 1;
				dd = 1;
			}
			break;
		case 'h': /* hours */
		case 'H':
			fd_temp = today + ((float)ndays * 3600l);
			if (fd_temp < 0)
				fd_temp = 1;
			if (fd_temp > maxint)
				today = maxint;
			else
				today = fd_temp;
			tm = localtime(&today);
			mm = tm->tm_mon + 1;
			dd = tm->tm_mday;
			yy = (int) tm->tm_year;
			hour = tm->tm_hour;
			min = tm->tm_min;
			sec = tm->tm_sec;
			break;
		case 'm':
		case 'M':
			nmonths = abs(ndays);
			k = nmonths / 12;
			yy = (int) tm->tm_year - 80 + (ndays < 0 ? 0 - k : k);
			nmonths -= k * 12;
			mm = tm->tm_mon + 1 + (ndays < 0 ? 0 - nmonths : nmonths);
			if (mm < 1) {
				mm += 12;
				--yy;
			}
			else if (mm > 12) {
				mm -= 12;
				++yy;
			}
			dd = tm->tm_mday;
			if (yy > 121) {
				fprintf(stderr,"Invalid Date.  Date will be set to 1/1/2021\n");
				yy = 121;
				mm = 1;
				dd = 1;
			}
			break;
		default:
			// ndays is actually a count of days 
			// we need to translate it to seconds
			// and add to the current time.
			// MAXDAYS1 is the largest possible number to not overflow
			offset = (float)ndays * 24 * 3600;
			fd_temp = today + offset;
			if (fd_temp < 0)
				fd_temp = 1;
			if (fd_temp > maxint)
				today = maxint;
			else
				today = fd_temp;
			tm = localtime(&today);
			mm = tm->tm_mon + 1;
			dd = tm->tm_mday;
			yy = (int) tm->tm_year;	
		}
		if (yy < 80 || today < 0 || tm == NULL) {
			fprintf(stderr,"You specified a date earlier than Jan. 1, 1980.  Using 1/1/80.\n");
			yy = 80;
			mm = 1;
			dd = 1;
		}
		cfidate = (yy << 9) | (mm << 5) | dd;
	}
	else {
		// Time is absolute
		//temp begins with a number, start month there
		mm = (temp[0]-'0');
		start = 2;
		if (isdigit(temp[1])) 
			mm = mm*10 + temp[1]-'0';
		
		if (isdigit(temp[start]))
			;
		else
			++start;

		dd = (temp[start]-'0');
		if (isdigit(temp[start+1])) {
			dd = dd*10 + temp[3]-'0';
			start += 2;
		}
		else
			start++;
		
		if (isdigit(temp[start]))
			;
		else
			++start;

		yy = (temp[start]-'0') *10 + temp[start+1]-'0';
		strcpy(tmpp,temp);  // Save 1st part of date
		if (temp[start+2] == 0 || temp[start+2] == ':' || temp[start+2] == '-') {
			//printf("%d %d %d\n",mm,dd,yy);
			// have a two digit year
			if (yy < 80)
				yy += 100;
			start += 2; // position start past the year
		}
		else  {
			/* If year is 4 digits ending in 0 or : assume full year not yyhh */
			yy *= 100;
			start += 2;
			yy += (temp[start]-'0') *10 + temp[start+1]-'0';
			start += 2;
			yy -= 1900;  // set yy to years since 1900
		}

		if (temp[start] == ':')
			++start;

		if (temp[start] && temp[start] == '-')
			tmpp[start] = 0;
		else 
			if (temp[start+2] && temp[start+2] == '-')
				tmpp[start+2] = 0;
			else
				if (temp[start+4] && temp[start+4] == '-')
					tmpp[start+4] = 0;

		if (tmpp[start]) { /* hour specified */
			min = 0;
			sec = 0;
			// If a time is specified, reset min,sec back to 0 from 59
			hour = (tmpp[start]-'0') *10 + tmpp[start+1] - '0';
			if (tmpp[start+2]) { /* min specified */
				min = (tmpp[start+2]-'0') * 10 + tmpp[start+3] - '0';
				if (tmpp[start+4]) { /* sec specified */
					sec = (tmpp[start+4]-'0') *10 + tmpp[start+5] - '0';
				}
			}
		}
		//printf("%d %d %d\n",hour,min,sec);
		/* year must be relative 1980 */
		// if (yy < 0) {
		//	fprintf(stderr,"You specified a date earlier than Jan. 1, 1980.  Using 1/1/80.\n");
		//	yy = 0;
		//	mm = 1;
		//	dd = 1;
		//}
 /*		cfidate = (yy << 9) | (mm << 5) | dd; */		
	}
	sprintf(actdate,"%d/%d/%d %d:%d:%d",mm,dd,yy+1900,hour,min,sec);
	gooddate = 0;
	if (mm >=1 && mm <= 12)
		if (dd >= 1 && dd <= days[mm])
			if (yy >= 0)
				if (hour >= 0 && hour <= 23)
					if (min >= 0 && min <= 59)
						if (sec >= 0 && sec <= 59)
							gooddate = 1;

	if (gooddate == 0)
		return(-1l);

	//printf("%s\n",actdate);
	fitm.tm_year = yy; /* Years since 1900 */
	fitm.tm_mday = dd;
	fitm.tm_mon = mm - 1;
	fitm.tm_hour = hour;
	fitm.tm_isdst = -1;
	if (fitm.tm_hour > 23)
		fitm.tm_hour = 23;
	fitm.tm_sec =  sec; /* seconds in 2 second intervals */
	if (fitm.tm_sec > 59)
		fitm.tm_sec = 59;
	fitm.tm_min = min;
	if (fitm.tm_min > 59)
		fitm.tm_min = 59;
	fitm.tm_wday = 0;
	fitm.tm_yday = 0;
   	/* convert to seconds since 1970 */
	rc = mktime(&fitm);
//	if (rc == -1) 
//		fprintf(stderr, "Invalid date.\n");
	return rc;
}

/* this routine returns 1 if the path is a directory, else 0 */
int is_dir(char *path) {
	HANDLE path_handle;
	WIN32_FIND_DATA dta;
	char temp[_MAX_PATH],pth[_MAX_PATH];
	int fnd,isdir;
	strcpy(temp,path);
	isdir = 0;
	if (temp[1] == ':' && temp[2] == 0)
		strcat(temp,"\\");	
	_fullpath(pth,temp,_MAX_PATH);
	if ((fnd = stat(pth,&filestat)) == 0)
		isdir = (filestat.st_mode & S_IFDIR);
	else if (strncmp(pth,"\\\\",2))
		;
	else {
		strcpy(temp,pth);
		strcat(temp,"\\*.*");
		path_handle = INVALID_HANDLE_VALUE;
		path_handle = FindFirstFile(temp,&dta);
		if (path_handle != INVALID_HANDLE_VALUE) {
			isdir = 1;
			FindClose(path_handle);
		}
	}
	return isdir;
}

void usage(void) {
	char tmp[_MAX_PATH],*p,cmd[_MAX_PATH];
	FILE *hlp;

	if ((p=getenv("tmp")) == NULL)
		if ((p = getenv("temp")) == NULL)
			GetWindowsDirectory(tmp,_MAX_PATH);
		else
			strcpy(tmp,p);
	else
		strcpy(tmp,p);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");

	strcat(tmp,"search32.hlp");

	hlp = fopen(tmp,"w");

	fprintf(hlp,"Disk Search Filter Copyright London Computing, 1985-2003, (WIN32) %s.\n",__DATE__);
	fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
	fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(hlp,"(at your option) any later version.\n\n");
	fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
	fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(hlp,"GNU General Public License for more details.\n\n");
	fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
	fprintf(hlp,"along with this program; if not, write to the Free Software\n");
	fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");

	// fprintf(hlp,"Now Year 2000 Compliant.\n");
	fprintf(hlp,"Usage is:\nSearch32 [[path] ... file.ext ...] ...] [-FLAGS]\n");
	fprintf(hlp,"Flags:  [-F] [-H] [-A] [{-O=otfi|-L=otfi}] [-D[mdy1][-mdy2]] [-S[low][-high]\n");
	fprintf(hlp,"        [-R{+|-}] [-V[-]] [-B] [-T[O]] [-8] [-X=excludepath] [-Q] [-SC]\n");
	fprintf(hlp,"        [-I=\"string in file\"] [-P[t]] [-C] [-E] [-STREAMS] [-S+] [-S-]\n");
	fprintf(hlp,"\nScans all directories for file.ext ...  Multiple names may be specified.\n");
	fprintf(hlp,"Unix style wildcards are permitted in file names.  File names should NOT\n");
	fprintf(hlp,"include path or drive.  \n\n");
	fprintf(hlp,"Path names must include at least one backslash (\\).  Path names may include\n");
	fprintf(hlp,"drive letters.  Path names may	appear intermixed with file names.  Path\n");
	fprintf(hlp,"names should NOT include file names.\n\n");
	fprintf(hlp,"When present, a path name limits searching for all subsequent file\n");
	fprintf(hlp,"names to the specified drive and directory until a new directory name is\n");
	fprintf(hlp,"specified.  If more than one path precedes a list of files, all paths are\n");
	fprintf(hlp,"searched for the specified files.  If you do not specify a path name, \n");
	fprintf(hlp,"the default drive is searched\n\n");
	fprintf(hlp,"You may use ? to indicate all fixed drives and * to indicate\n");
	fprintf(hlp,"all fixed, CDROM and LAN drives.  If * or ? is used, no other paths can be\n");
	fprintf(hlp,"specified.  When * is used, SEARCH32 displays the UNC name of LAN paths.\n\n");
	fprintf(hlp,"Search32 returns 0 if any files found, 1 if no files found, else error.\n\n");
	fprintf(hlp,"If only one parameter is specified and it begins with an @, eg @xxx.inp, \n");
	fprintf(hlp,"then xxx.inp is read for all parameters.  Parameters should be placed one\n");
	fprintf(hlp,"per line.  A parameter could be a path, file or flag.\n");
	fprintf(hlp,"\nSEARCH32 Flags:\n");
	fprintf(hlp,"\n-F -- output a list of pathnames to each file found.  If omitted, \n");
	fprintf(hlp,"       output report showing file name, size, creation date and version.\n");
	fprintf(hlp,"\n-P -- Path flag.  Displays long file names as full path names.\n");
	fprintf(hlp,"-Pt-- Path flag.  Displays long file names as full path names, seperates\n");
	fprintf(hlp,"      fields with tab characters instead of spaces, places a tab character\n");
	fprintf(hlp,"      between the file path and file name fields.  Hard to read, easy\n");
	fprintf(hlp,"      to parse or import into a spreadsheet or database.  Omits ver: before\n");
	fprintf(hlp,"      version information.  Includes all fields (See -V+).\n"); 
	fprintf(hlp,"\n-H -- Hidden/System files only.\n");
	fprintf(hlp,"\n-A -- Archived files only.\n");
	fprintf(hlp,"\n-B -- Do not display file version information.  This is a workaround for a bug\n");
	fprintf(hlp,"       in the WIN32 API that causes abends on certain files when attempting to\n");
	fprintf(hlp,"       retrieve Windows version information.\n");
	fprintf(hlp,"\n-C -- Calculate a CRC for each file found.  If specified, CRC appears as the\n");
	fprintf(hlp,"       first column in the report replacing the short (8.3) filename.\n");
	fprintf(hlp,"\n-Q -- Do not display Copyright Notice and starting path for Search.\n");
	fprintf(hlp,"\n-Ootfi -- Send output to file otfi.\n");
	fprintf(hlp,"\n-Lotfi -- Log output to file otfi and display on screen.\n");
	fprintf(hlp,"\n-W-  Return the earliest Write date, i.e. oldest file found.\n");
	fprintf(hlp,"-W+    Return the newest file found.\n");
	fprintf(hlp,"-W[mdy1][mdy2] Same as -D[mdy1][mdy2] below.\n");
	fprintf(hlp,"\n-D-   Return the oldest file found.\n");
	fprintf(hlp,"-D+   Return the newest file found.\n");
	fprintf(hlp,"\n-D[mdy1][-mdy2] -- Only select files with date of last write between mdy1-mdy2\n");
	fprintf(hlp,"        Specify dates as mmddyy[yy][:hh[mm[ss]]].  If mdy1 omitted, look for\n");
	fprintf(hlp,"        dates <= mdy2.  If mdy2 omitted, look for dates >= mdy1.\n\n");
	fprintf(hlp,"        You may specify an optional file time along with the file date.\n");
	fprintf(hlp,"        The file time is specified in hours (hh), optional minutes (mm)\n");
	fprintf(hlp,"        and optional seconds (ss).  If you do not specify a time, then\n");
	fprintf(hlp,"        mdy1 assumes a time of 00:00:00 and mdy2 assumes a time of 23:59:59.\n\n");
	fprintf(hlp,"        You may use () to indicate todays date.  Use (n) for todays date\n");
	fprintf(hlp,"        plus n days, (nM) for todays date plus n months, (nY) for todays\n");
	fprintf(hlp,"        plus n years.  Use (-n) for todays date minus n days, (-nM) for\n");
	fprintf(hlp,"        todays date -n months, (-nY) for todays date minus n years.\n");
	fprintf(hlp,"        Use (-nH) for current date/time minus n hours.\n\n");
	fprintf(hlp,"        Any date earlier than 1/1/1980 is set to 1/1/1980.\n");
	fprintf(hlp,"        Dates with 2 digit years >= 80 are assumed to be relative to 1900.\n");
	fprintf(hlp,"        Dates with 2 digit years < 80 are assumed to be relative to 2000.\n");
	fprintf(hlp,"        If years are 4 positions, they are assumed to be exact. \n");
	fprintf(hlp,"\n-A[mdy1][-mdy2] -- Only select files with date of last access between mdy1-mdy2\n");
	fprintf(hlp,"        Specify dates as mmddyy[yy][:hh[mm[ss]]].  If mdy1 omitted, look for\n");
	fprintf(hlp,"        dates <= mdy2.  If mdy2 omitted, look for dates >= mdy1.\n\n");
	fprintf(hlp,"        You may specify an optional file time along with the file date.\n");
	fprintf(hlp,"        The file time is specified in hours (hh), optional minutes (mm)\n");
	fprintf(hlp,"        and optional seconds (ss).  If you do not specify a time, then\n");
	fprintf(hlp,"        mdy1 assumes a time of 00:00:00 and mdy2 assumes a time of 23:59:59.\n\n");
	fprintf(hlp,"        You may use () to indicate todays date.  Use (n) for todays date\n");
	fprintf(hlp,"        plus n days, (nM) for todays date plus n months, (nY) for todays\n");
	fprintf(hlp,"        plus n years.  Use (-n) for todays date minus n days, (-nM) for\n");
	fprintf(hlp,"        todays date -n months, (-nY) for todays date minus n years.\n");
	fprintf(hlp,"        Use (-nH) for current date/time minus n hours.\n\n");
	fprintf(hlp,"        Any date earlier than 1/1/1980 is set to 1/1/1980.\n");
	fprintf(hlp,"        Dates with 2 digit years >= 80 are assumed to be relative to 1900.\n");
	fprintf(hlp,"        Dates with 2 digit years < 80 are assumed to be relative to 2000.\n");
	fprintf(hlp,"        If years are 4 positions, they are assumed to be exact. \n");
	fprintf(hlp,"        When specified, the last activity date will be displayed instead of \n");
	fprintf(hlp,"        last write date.\n");
	fprintf(hlp,"\n-I=\"string in file\"  \n");
	fprintf(hlp,"      Select the file only if it contains the string.  Enclose the string\n");
	fprintf(hlp,"      in quotes if it contains spaces or special characters.  Case ignored.\n");
	fprintf(hlp,"      Searching for a string will significantly increase the time of the \n");
	fprintf(hlp,"      search.  The search string may be up to 255 characters.\n\n");
	fprintf(hlp,"      You may specify up to 16 -I strings.  If more than one is present\n");
	fprintf(hlp,"      the file will be returned if any of the strings are present.\n");
	fprintf(hlp,"\nt-Streams  If present, display all the streams of selected files.\n");
	fprintf(hlp,"\n-S-   Return the smallest file found.\n");
	fprintf(hlp,"-S+   Return the largest file found.\n");
	fprintf(hlp,"-SC   Write all file sizes using commas.\n");
	fprintf(hlp,"\n-S[low][-high] -- Only select files with file size between low-high\n");
	fprintf(hlp,"        Specify low, high in bytes.  If low omitted, look for sizes <= high.\n");
	fprintf(hlp,"        If high omitted, look for sizes >= low.\n");
	fprintf(hlp,"\n-R- --  Turn off recursion.  Only look in specified directory.\n");
	fprintf(hlp,"-R+ --  Turn on recursion.  Look in all subdirectories (default).\n");
	fprintf(hlp,"\n-T  --  Display number of files found, total file size of files found,\n");
	fprintf(hlp,"        and the total space used by these files on disk.\n");
	fprintf(hlp,"\n-TO --  As above, but does not display a list of files found. \n");
	fprintf(hlp,"\n-V  --  Verbose mode.  List names of Directories Being Searched,\n");
	fprintf(hlp,"        all file dates and file owner.  Default is on\n");
	fprintf(hlp,"-V- --  Turn off Verbose Mode.  Don't display directories being searched.\n");
	fprintf(hlp,"-V+ --  Enhanced Verbose mode, display all known file attributes including \n");
	fprintf(hlp,"        Last Access Date, Creation Date, File Owner, MachineName and file attributes.\n");
	fprintf(hlp,"\n-8  --  Show 8.3 file names only;    default for DOS, WIN 3.1 and OS2.\n");
	fprintf(hlp,"\n-E  --  Examine each file selected to determine if the file contains\n");
	fprintf(hlp,"        multiple streams.  If so, look in all streams.\n");
	fprintf(hlp,"\n-1  --  Return the first file that matches the specifications.\n");
#ifdef UNIX_EXCLUDES
	fprintf(hlp,"\n-X=str--Do not report on any file if its fully qualified file name\n");
	fprintf(hlp,"        contains the string str.  You may specify up to 255 seperate\n");
	fprintf(hlp,"        -X statements.  If the filename matches any of the exclude \n");
	fprintf(hlp,"        strings, it will be excluded.  Enclose the string in quotes \n");
	fprintf(hlp,"        if it contains any special characters or white space.\n");
	fprintf(hlp,"        You may use Unix style wildcards in any of the exclude strings.\n");
	fprintf(hlp,"        When using Unix style wildcards, you must use two \\\\ to specify\n");
	fprintf(hlp,"        a single \\ in the string.  See below.\n");
	fprintf(hlp,"        Case is ignored when searching for exclude strings.\n");
#else
	fprintf(hlp,"\n-X=str--Exclude any files or subdirectories that contain the string str.\n");
	fprintf(hlp,"        You may specify up to 255 seperate -X statements.  A subdirectory\n");
	fprintf(hlp,"        or file will be excluded if it contains any string from any -X statement.\n");
	fprintf(hlp,"        Case is ignored when searching for exclude strings.  You must use two \\\\\n");
	fprintf(hlp,"        for every \\ that precedes a \" character within a quoted string.\n");
	fprintf(hlp,"        If you forget, the ending \" will become part of the exclude string and you\n");
	fprintf(hlp,"        will see some very weird results.\n");
#endif
	fprintf(hlp,"\n-X=@std If you specify @std for the exclude string, a standard set of\n");
	fprintf(hlp,"        excludes will be generated consisting of:\n");
	fprintf(hlp,"        \\TEMP\\, TEMPORARY, $NtUninstall, COOKIES, .LNK and .TMP.  This \n");
	fprintf(hlp,"        will speed up your search, but may exclude some directories that you\n");
	fprintf(hlp,"        might want to scan.\n");
	fprintf(hlp,"\n-M=MsgFile -- If any files are found, prefix the MsgFile to the output.\n");
	fprintf(hlp,"\nUnix Style Wildcards:\n");
	fprintf(hlp,"Unix Wildcards may be used in any file name being searched for:\n");
	fprintf(hlp,"The wildcard symbols consist of *, ? and [...].\n");
	fprintf(hlp,"\n? is a place holder for 0 or 1 characters.  ? will match any character.\n");
	fprintf(hlp,"\n* is a place holder for any string of characters.  Unix allows more than\n");
	fprintf(hlp,"one * in a filename,  e.g.  *xyz* stands for any filename containing the\n");
	fprintf(hlp,"letters xyz anywhere in the name.  * also matches the period between the\n");
	fprintf(hlp,"filename and filetype in a DOS filename, e.g.  scan* will match SCAN.EXE\n");
	fprintf(hlp,"as well as SCANMAN or SCANMAN.BAT.\n");
	fprintf(hlp,"\n[...] allows you to look for a specific set of characters in any position\n");
	fprintf(hlp,"of a filename.  LOOK[ACG-K].EXE will match LOOKA.EXE, LOOKC.EXE,\n");
	fprintf(hlp,"LOOKG.EXE, LOOKH.EXE, LOOKI.EXE, LOOKJ.EXE and LOOKK.EXE.  You may place\n");
	fprintf(hlp,"individual letters within [...] or a range of letters.  Ranges are\n");
	fprintf(hlp,"specified as X-Y where X is the first letter in the range and Y is the last\n");
	fprintf(hlp,"\nExamples:\n");
	fprintf(hlp,"\nSearch for all .BAK files and output a list of files to bakfiles.dat\n");
	fprintf(hlp,"SEARCH32 *.BAK -f -obakfiles.dat    -OR- SEARCH4 *.BAK -f >bakfiles.dat\n\n");
	fprintf(hlp,"\nSearch for all .BAK files in current directory\n");
	fprintf(hlp,"SEARCH32 *.BAK -f -r-\n\n");
	fprintf(hlp,"\nSearch for all .C files in the C:\\MSC or the D:\\BORLANDC directories\n");
	fprintf(hlp,"SEARCH32 C:\\MSC *.C D:\\BORLANDC *.C -OR- SEARCH4 C:\\MSC D:\\BORLANDC *.C\n\n");
	fprintf(hlp,"\nSearch for all files no more than 180 days old on drives C:, D: and E:\n");
	fprintf(hlp,"SEARCH32 C: D: E: * -d(-180)\n\n");
	fprintf(hlp,"\nSearch for all files at least 180 days old on drives C:, D: and E:\n");
	fprintf(hlp,"SEARCH32 C: D: E: * -d-(-180)\n\n");
	fprintf(hlp,"\nSearch for all files between 15 and 18 months old on the current drive.\n");
	fprintf(hlp,"SEARCH32 \\ * -d(-18m)-(-15m)\n\n");
	fprintf(hlp,"\nSearch for all files less than 7 days old that are larger than 1MB on drive C:.\n");
	fprintf(hlp,"SEARCH32 C: * -d(-7d) -s1000000\n\n");
	fprintf(hlp,"\nSEARCH for all files on drive C less 33 bytes long.\n");
	fprintf(hlp,"SEARCH32 C: * -s-32\n");
	fprintf(hlp,"\nSearch for all .DLL files on local drives.\n");
	fprintf(hlp,"SEARCH32 ? *.DLL\n");
	fprintf(hlp,"\nSearch for all .VBX files on any accessible drive.\n");
	fprintf(hlp,"SEARCH32 * *.VBX\n");
	fprintf(hlp,"\nSearch for all on a local drive files dated between\n01/01/1998 13:00:00 and 01/02/1998 11:55:00\n");
	fprintf(hlp,"SEARCH32 ? * -d010198:1300-010298:1155  -or-\n");
	fprintf(hlp,"SEARCH32 ? * -d01011998:1300-01021998:1155\n");
#ifdef UNIX_EXCLUDES
	fprintf(hlp,"\nSearch for all files on the local drives.  Exclude any files of type\n");
	fprintf(hlp,".OBJ or .TMP\n");
	fprintf(hlp,"SEARCH32 ? * -X=\"*.TMP\" -X=\"*.OBJ\"\n");
	fprintf(hlp,"\nSearch for all files on the local drives.  Exclude any files in any\n");
	fprintf(hlp,"subdirectories starting with TEMP or TMP.  The patterns used below look\n");
	fprintf(hlp,"for \\TMP...\\ and \\TEMP...\\ and exclude if found.  Remember that Unix\n");
	fprintf(hlp,"Wild Card processing uses \\ as an escape character.  That is why the \n");
	fprintf(hlp,"exclude path specifies \\\\. \\\\ must be used for a single \\.\n");
	fprintf(hlp,"SEARCH32 ? * -X=\"*\\\\TMP*\\\\*\" -X=\"*\\\\TEMP*\\\\*\"\n");
#else
	fprintf(hlp,"\nSearch all files on the local drives.  Exclude any files of type\n");
	fprintf(hlp,".OBJ or .TMP\n");
	fprintf(hlp,"SEARCH32 ? * -X=\".TMP\" -X=\".OBJ\"\n");
	fprintf(hlp,"\nSearch all files on the local drives.  Exclude any files in \n");
	fprintf(hlp,"any subdirectories starting with TEMP or TMP.  The patterns used\n");
	fprintf(hlp,"below look for \\TMP... and \\TEMP... and exclude if found.  The\n");
	fprintf(hlp,"example below will also exclude any files starting with TMP or TEMP.\n");
	fprintf(hlp,"SEARCH32 ? * -X=\"\\TMP\" -X=\"\\TEMP\"\n");
	fprintf(hlp,"\nSearch all files on the local drives.  Exclude any files in \n");
	fprintf(hlp,"in subdirectories starting with TEMP, TMP or Temporary Internet Files.\n");
	fprintf(hlp,"The patterns used below look for \\TMP\\, \\TEMP\\ and \n");
	fprintf(hlp,"\\Temporary Internet Files\\ and exclude if found.  This example only\n");
	fprintf(hlp,"excludes directories, not files because the strings start and end with \\.\n");
	fprintf(hlp,"SEARCH32 ? * -X=\"\\TMP\\\" -X=\"\\TEMP\\\" -X=\"\\Temporary Internet Files\\\"\n");
#endif
	fclose(hlp);

	strcpy(cmd,"NOTEPAD.EXE ");
	strcat(cmd,tmp);

	WinExec(cmd,SW_SHOW);

	exit(1);
}



void setDateRange(char *arg, char *earliest_date, char *latest_date, time_t *cfidate, 
			time_t *cfidate_hi, char *msg) {
	char temp[255], early[25], late[25], tmp[255];
	char *minus;
	time_t t;

		/* have a date range argument of form */
		if (arg[2] == '-' && arg[3] == 0) {
			*earliest_date = 1;
			fprintf(stderr,"Searching for oldest %s.\n",msg);
			return;
		}
				
		if (arg[2] == '+' && arg[3] == 0) {
			*latest_date = 1;
			fprintf(stderr,"Searching for newest %s.\n",msg);
			return;
		}
		/* -D[mmddyy][-mmddyy] ===> select files with */
		/* filedate >= first date and <= second date */
		if (arg[2] == '-') {
			strcpy(temp,arg+3);
			t = getadate(temp,late,'H');
			if (t == -1) {
				fprintf(stderr,"Invalid Date: %s\n",temp);
				exit(1);
			}
			else {
				*cfidate_hi = t;
				fprintf(stderr,"Searching for %s earlier than %s\n",msg,late);
			}
		}
		else {
		// copy date or date range into temp
			strcpy(temp,arg+2);
			t = getadate(temp,early,' ');	//get low date
			if (t == -1) {
				fprintf(stderr,"Invalid Date: %s\n",temp);
				exit(1);
			}
			else
				*cfidate = t;
				
			// printf("%s\n",early);
			minus = temp; //look for ( in date.  If found, skip past )
			if (*minus == '(') {
				minus = strchr(temp,')');
				strcpy(temp,minus+1);
			}
			minus = strchr(temp,'-');
			// printf("%s\n",early);
			if (minus) {
				/* The release level compiler resets early */
				/* The debug does not.  The following statement */
				/* forces the correct behaviour */
				strcpy(tmp,early);
				strcpy(temp,minus+1);
				t = getadate(temp,late,'H');
				if (t == -1) {
					fprintf(stderr,"Invalid Date: %s\n",temp);
					exit(1);
				}
				else
					*cfidate_hi = t;
				strcpy(early,tmp);
				fprintf(stderr,"Searching for %s between %s and %s\n",msg,early,late);
			}
			else
				fprintf(stderr,"Searching for %s later than %s\n",msg, early);
		}
}


FILE *msgfi;
void setflag(char *arg,char *logname) {
	int o;
	time_t t;
	char temp[_MAX_PATH],early[16],late[16];
	char *minus, tmp[_MAX_PATH];
// Check the flag value and set a global variable

	switch (toupper(arg[1])) {
	
	case '1':
		first_file = 1;
		break;

	case 'I':
		if (search_string_count < 16) {
			o=2;
			if (arg[o] == '=' || arg[o] == ' ')
				++o;
			if (strlen(arg) > 256)
				fprintf(stderr,"%s\nSearch string exceeds 255 characters.\n",arg);
			else {
				strcpy(search_string[search_string_count++],arg+o);
				if (search_string_count > 16) 
					fprintf(stderr,"You have specified more than 16 search strings.\n");
			}	
		}
		break;

	case 'M':  /* Message to Prefix Output */
		o=2;
		if (arg[o] == '=' || arg[o] == ' ')
			++o;
		if ((msgfi = fopen(arg+o,"r")) == NULL) {
			fprintf(stderr,"Unable to Open %s\n",arg+3);
			exit(1);
		}                                   
		break; 

	case 'L':  /* Log to a File */
		logfile = 1;
		o=2;
		if (arg[o] == '=' || arg[o] == ' ')
			++o;
		strcpy(logname, arg+o);
		if ((logfi = fopen(arg+o,"w")) == NULL) {
			fprintf(stderr,"Unable to Open %s\n",arg+3);
			exit(1);
		}     
		break;

	case 'B':
		display_version_info = 0; // Omit version info
		break;

	case 'Q': // Quiet flag
		quiet = 1;
		break;
	
	case 'X': // Exclude files
		o=2;
		if (arg[o] == '=' || arg[o] == ' ')
			++o;
		if (stricmp(arg+o,"@std")==0) {
			char *token;
			// Remove 'standard' paths that contain temporary variables,
			// Internet Files, Links, etc.
			token = strtok( notempexcludes, "\n");
			while(token != NULL && exclude_string_count < MAX_EXCLUDES)  {
				/* While there are tokens in "string" */
				exclude_path[exclude_string_count] = strupr(token);
				fprintf(stderr,"Exclude: %s\n",exclude_path[exclude_string_count]);
				++exclude_string_count;
				/* Get next token: */
				token = strtok( NULL, "\n" );
			}
		}
		else if (strlen(arg) > _MAX_PATH)
			fprintf(stderr,"%s\nSearch string exceeds %d characters.\n",arg,_MAX_PATH);
		else {
			if (exclude_string_count < MAX_EXCLUDES) {
				exclude_path[exclude_string_count] = (char *) malloc(strlen(arg)+1);
				if (exclude_path[exclude_string_count] == NULL) {
					fprintf(stderr,"Out of Memory saving Exclude Strings.\n");
					printf("Out of Memory saving Exclude Strings.  %s has been ignored.\n");
				}
				else {
					strcpy(exclude_path[exclude_string_count],arg+o);
					strupr(exclude_path[exclude_string_count]);
					fprintf(stderr,"Exclude: %s\n",exclude_path[exclude_string_count]);
					++exclude_string_count;
				}
			}
			else
				fprintf(stderr,"Too many exclude strings.  %s has been ignored.\n",arg);
		}
		break;
	
	case '8': // use 8.3 names everywhere
		short_name = 1;
		break;
		
	case 'C': // calculate a CRC for each file found
		calculate_crc = 1;
		init_crc32();
		break;
	
	case 'R': // Recurse + (yes -- default); - (no)
		recursion = (arg[2] == '+' ? 1 : 0);
		break;
		
	case 'P': // Use full path name for long file name
		pathflag = 1;
		if (toupper(arg[2]) == 'T') {
			tabs = 1;
			space_or_tab[0] = '\t';
		}
		break; // output statistics followed by full file path
		
	case 'F': // output filename only
		fileform = 1;
		break; /* output as filename */
		
	case 'A': // Only look for files with the archive bit or look at active dates stamps
		if (arg[2]==0)
			archived = 1;
		else
			setDateRange(arg, &earliest_activity, &latest_activity, &afidate, 
			&afidate_hi, "files with access date");
		break; /* done processing -A flag*/

	
	case 'T': // Display total size and number of files found; -to show only totals
		tflag = 1; /* report totals */
		totalonly = (toupper(arg[2]) == 'O' ? 1 : 0);
		break;
		
	case 'H': // Only show hidden or system files
		hidden_system = 1;
		break; /* hidden/system files only */
		
	case 'O':  /* output to a file */
		o=2;
		if (arg[o] == '=' || arg[o] == ' ')
			++o;
		if ((otfi = fopen(arg+o,"w")) == NULL) {
			fprintf(stderr,"Unable to Open %s\n",arg+o);
			exit(1);
		}
		break;
		
	case 'D': // Select within a date range or only oldest or newest file
		setDateRange(arg, &earliest_date, &latest_date, &cfidate, 
		&cfidate_hi, "files with change date");
		break; /* done processing -A flag*/

	case 'V':	// Verbose mode.  Show the directories that are being searched
		vflag = (arg[2]=='-' ? 0 : 1);
		vplus = (arg[2]=='+' ? 1 : 0);
		break;
		
	case 'S':   
		// Look for -SC, displays all sizes using commas
		if (stricmp(arg,"-SC") == 0) {
			use_commas = 1;
			break;
		}
			
		// Look for -STREAMS -or- specify a range of files sizes to look for -or-
		// return the largest or smallest files
		// -STREAMS -or- -slow-high -or- -s- -or- -s+
		if (strnicmp(arg,"-ST",3) == 0) {
			// Have Streams
			checkStreams = 1;
			break;
		}
		/* have a size range argument of form */
		/* -D[size1][-size2] ===> select files with */
		/* size >= size1 and <= size2 */
		if (arg[2] == '-' && arg[3] == 0) {
			smallest_file = 1;
			fprintf(stderr,"Searching for smallest file.\n");
			break;
		}
				
		if (arg[2] == '+' && arg[3] == 0) {
			largest_file = 1;
			fprintf(stderr,"Searching for largest file.\n");
			break;
		}
				
		if (arg[2] == '-') {
			sizehigh = atol(arg+3);
			if (sizehigh == 0)
				sizehigh = 0x7fffffff;
			fprintf(stderr,"Searching for file sizes <= %I64d\n",sizehigh);
		}
		else {
			strcpy(temp,arg+2);
			minus = strchr(temp,'-');
			if (minus)
				*minus = 0;
			sizelow = atol(temp);
			if (minus) {
				sizehigh = atol(minus+1);
				if (sizehigh == 0)
					sizehigh = 0x7fffffff;
				fprintf(stderr,"Searching for file sizes >= %I64d and <= %I64d\n",sizelow,sizehigh);
			}
			else
				fprintf(stderr,"Searching for file sizes >= %I64d\n",sizelow);
		}
		break;
	}
}


unsigned GetDriveType16(int i) {
	char drive[8];
	strcpy(drive,"C:\\");
	drive[0] = 'A'+i;
	return GetDriveType(drive);
}


int StartLooking(char *args[256], int k, int startfile, int endfile) {
	int i, j, tp, bufSizeNeeded, li, lj;
	char temp[_MAX_PATH], path[_MAX_PATH];

	// Special case, a ? stands for all internal drives (DRIVE_FIXED)
	if (args[k][0] == '?' && args[k][1] == 0) {
		for (i=2;i < 26; ++i) {
			if ((tp=GetDriveType16(i)) == DRIVE_FIXED) {
				strcpy(temp,".:\\");
				temp[0] = 'A' + i;           
				strcpy(path,temp);
				strupr(path);
				look4(args,startfile+1,endfile,path);
			}
		} 
		return endfile;
	}
	
	// Special case, a * stands for all files, but try not to search any network
	// path twice, even if it was mapped to two different drive letters
	if (args[k][0] == '*' && args[k][1] == 0) {
		// Determine if any net connection is a subset or
		// duplicate of any other and skip it
		for (i=2; i < 26; ++i) {
			net_connection[i][0] = 0;
			if ((tp = GetDriveType16(i)) == DRIVE_REMOTE) {
				temp[0] = 'A' + i;
				temp[1] = ':';
				temp[2] = 0;
				bufSizeNeeded = _MAX_PATH;
				if (WNetGetConnection(temp, net_connection[i], &bufSizeNeeded) == NO_ERROR) {
					strupr(net_connection[i]);
					li = strlen(net_connection[i]);
					for (j=2; j < i && net_connection[i][0]; ++j) {
						if (net_connection[j][0] == 0)
							continue;
						// look for i in j or j in i and remove
						// longer version
						lj = strlen(net_connection[j]);
						// if both are the same, don't record i
						if (strcmp(net_connection[i],net_connection[j]) == 0) {
							net_connection[i][0] = 0;
							continue;
						}
						// if i is longer than j and j is a subset of i and
						// next position of i is \; set i to empty
						if (li > lj) 
							if (strstr(net_connection[i], net_connection[j]) == net_connection[i] && net_connection[i][lj] == '\\')
								net_connection[i][0] = 0;
							else
								;
						else // they are the same size or j is shorter than i
							if (strstr(net_connection[j], net_connection[i]) == net_connection[j] && net_connection[j][li] == '\\')
								net_connection[j][0] = 0;
							else
								;
					}
				}
				else
					net_connection[i][0]= 0;
			}
		}
		
		for (i=2; i < 26; ++i) {
			// Create path to search
			strcpy(temp,".:\\");
			temp[0] = 'A' + i; 
			tp = GetDriveType16(i); // Get type of drive
			if (tp == 0)
				continue; /* If the drive is real, report on it */
			if (tp == DRIVE_NO_ROOT_DIR)
				continue; /* Drive letter is valid, but drive isn't there */
			// If a network drive, only report if not a dupe
			if (tp == DRIVE_REMOTE && net_connection[i][0] == 0)
				continue;
			// If we have a LAN connection, use UNC name
			if (tp == DRIVE_REMOTE) {
				strcpy(temp,net_connection[i]);
				if (temp[strlen(temp)-1] != '\\')
					strcat(temp,"\\");
			}
			strcpy(path,temp);
			strupr(path);
			// if (!quiet)	fprintf(stderr,"\nSearching %s ...\n",path);
			look4(args,startfile+1,endfile,path);
		}
		return endfile;
	}                                               
	
	if (k == startfile) {
		/* using current path */
		// if (!quiet)	fprintf(stderr,"\nSearching %s ...\n",path);
		look4(args,startfile,endfile,path);
		return endfile;
	}

	while (k < startfile) {
		strcpy(temp,args[k]);
		if (temp[1] == ':' && temp[2] == 0)
			strcat(temp,"\\");
		_fullpath(path,temp,_MAX_PATH);
		strupr(path);
		// if (!quiet)	fprintf(stderr,"\nSearching %s...\n",path);
		look4(args,startfile,endfile,path);
		++k;
	}
	return endfile;
}

// Display summary information and totals
void print_summary(char *logname) {
	char temp[_MAX_PATH], c;
	char tfiles[16],tused[16],tspace[16];

	temp[8] = 0;
	if (vflag)
		blankpathdisplay(temp+8);
	/* if VLAG, get rid of last path. */

	if (largest_file) 
		if (fileform)
			fprintf(otfi,"%s\n",large_file);
		else
			fprintf(otfi,"Largest  file: %s %s Bytes\n", large_file, comma_delim(largest_size,temp,0,0,1));

	if (smallest_file) 
		if (fileform)
			fprintf(otfi,"%s\n",small_file);
		else
			fprintf(otfi,"Smallest file: %s %s Bytes\n", small_file, comma_delim(smallest_size,temp,0,0,1));

	if (latest_date) 
		if (fileform)
			fprintf(otfi,"%s\n",late_file);
		else 
			fprintf(otfi,"Newest   file: %s %s",late_file, ctime(&late_date));

	if (earliest_date)
		if (fileform)
			fprintf(otfi,"%s\n",early_file);
		else 
			fprintf(otfi,"Oldest   file: %s %s",early_file, ctime(&early_date));

	if (tflag) {
		fprintf(otfi,"\nFound %s files, %s bytes in files, %s bytes on disk.\n",
		comma_delim(total_files,tfiles,0,0,1),
		comma_delim(total_space,tspace,0,0,1),
		comma_delim(total_used,tused,0,0,1));
		if (logfile)
			fprintf(logfi,"\nFound %s files, %s bytes in files, %s bytes on disk.\n",
			comma_delim(total_files,tfiles,0,0,1),
			comma_delim(total_space,tspace,0,0,1),
			comma_delim(total_used,tused,0,0,1));
	}

	// if msgfi specified, copy contents to otfi
	if (msgfi) {
		if (total_files) {
		// Copy contents of msgfi to otfi
		    while ((c = getc(msgfi)) != -1) {
				putc(c,otfi);    
				if (logfi)
					putc(c,logfi);
			}
		}                                               
		fclose(msgfi);
	}		
		
	// If there is a logfile, let use know and close it
	if (logfi) {
		fprintf(otfi,"\nResults have been logged to %s\n",logname);
		fclose(logfi);
	}		
}

int AddIndirectArgs(char *args[256], int argk) {
	char pth[_MAX_PATH], *argbuf;
	int i;
	FILE *fi;

	// If any arguments begin with a @, assume the rest of the argument is the name of
	// a file and read in switches from the file
	strcpy(pth,args[1]+1);
	fi = fopen(pth,"r");
	if (fi == NULL) {
		fprintf(stderr,"Unable to open %s\n",pth);
		exit(12);
	}

	i = fstat(_fileno(fi),&fistat);
	if (i == -1) {
		fprintf(stderr,"Unable to access file stats for %s\n",pth); 
		fclose(fi);
		exit(12);
	}

	argbuf = malloc((unsigned)fistat.st_size+6);
	if (argbuf == NULL) {
		fprintf(stderr,"Insufficient Memory.\n");
		fclose(fi);
		exit(12);
	}              

	fread(argbuf,1,(unsigned)fistat.st_size,fi);
	fclose(fi); 

	argbuf[fistat.st_size] = 0;
	for (argk=1, args[1]=argbuf, i=0; i < fistat.st_size; ++i) {
		if (argbuf[i] == '\n') {
			argbuf[i] = 0;
			++argk;
			args[argk] = argbuf+i+1;
		}
	}  
	args[argk] = 0;
	return argk;
}


void main(int argc,char *argv[])
{
	int lastfile,k,startfile,endfile,pathsize;
	int driveno,rc,isdir,i;
	char *args[256], logname[_MAX_PATH];
	char path[_MAX_PATH];
	int argk;

	logfi = NULL;
	msgfi = NULL;

	otfi = stdout;
	pathsize = sizeof(path);
	cfidate = 0;
	hidden_system = 0;
	archived = 0;
	cfidate_hi = 0x7FFFFFFF;
	afidate_hi = 0x7FFFFFFF;
	sizelow = 0;
	sizehigh = 0x7fffffffffffffff;  // set to a very large number 
	
	_tzset();
	if (argc < 2) {
		usage(); // display help
		exit(1);
	}
     
	getcwd(path,sizeof(path));
	driveno = path[0] - '@';
	path[1]=':';
	path[2]='\\';
	path[3]=0;
	lastfile = argc - 1;
	fileform = 0;
	recursion = 1;

	/* Set args to input arguments */
	argk = argc;		
	for (i=0; i <= argc; ++i)
		args[i] = argv[i];
	
	// Look for argument 1.  If it starts with an @, assume it points to a file containing
	// arguments to be used.

	if (argc == 2 && argv[1][0] == '@') 
		argk = AddIndirectArgs(args,argk);

//	Evaluate the argument list and set flags
	while (args[argk-1][0] == '-' && argk > 0) {
		setflag(args[argk-1],logname);
		argk--;
	}
	lastfile = argk-1;
		
//	Display the include strings
	if (search_string_count) {
		fprintf(stderr,"Searching for files containing any of the following:\n");
		for (k=0; k < search_string_count; ++k)
			fprintf(stderr,"\t%s\n",search_string[k]);
	}
//	Display the copyright notice
	if (!quiet) {
		fprintf(stderr,"Disk Search Filter Copyright London Computing, 1985-2002, (WIN32) %s.\n\n",__DATE__);
		fprintf(stderr,"");
	}

	for (k=1;k<=lastfile;++k) {
		/* for all file arguments, decide if directory or file */
		startfile = k;
		rc = 0;
	  /* look up entry in directory */
		isdir=is_dir(args[k]);
		if (isdir) {
			/* entry on command line is a directory */
			/* look for the first file entry after a list of directories */
			while (rc == 0 && startfile <= lastfile && isdir)
				isdir = is_dir(args[++startfile]);
		}
		/* startfile now set to the first file in the list */
		/* look for the last file argument before a new path */
		endfile = startfile;
		while (endfile < lastfile && !isdir) {
			isdir = is_dir(args[++endfile]);
		}
		if (endfile < lastfile)
			--endfile;
   		/* have 0 or more directories followed by 1 or more filespecs */

 		k = StartLooking(args, k, startfile, endfile);
	}	

	
	print_summary(logname);

	if (total_files)
		exit(0);
	else
		exit(1);
}
